@@ -1,4 +1,4 @@
-/* Copyright © 2015-2019 Pascal JEAN, All rights reserved.
+/* Copyright (c) 2015-2023 Pascal JEAN, All rights reserved.
  *
  * mbpoll is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -30,8 +30,13 @@
 #include <stdbool.h>
 #include <stdint.h>
 #include <time.h>
+#if defined (__unix__) || (defined (__APPLE__) && defined (__MACH__))
 #include <unistd.h>
 #include <sys/time.h>
+#endif
+#ifdef _WIN32
+#include <windows.h>
+#endif
 #include "serial.h"
 #include "custom-rts.h"
 #include "version-git.h"
@@ -50,7 +55,7 @@
 /* macros =================================================================== */
 #define BASENAME(f) (f)
 
-#if !defined(_WIN32) && (defined(__unix__) || defined(__unix) || (defined(__APPLE__) && defined(__MACH__)))
+#if defined (__unix__) || (defined (__APPLE__) && defined (__MACH__))
 # include <libgen.h>
 # undef BASENAME
 # define BASENAME(f) basename(f)
@@ -62,6 +67,12 @@
 #define PDEBUG(...)  if (ctx.bIsVerbose) printf(__VA_ARGS__)
 #endif
 
+#ifndef MIN
+#define MIN(a,b) (((a)<(b))?(a):(b))
+#endif
+#ifndef MAX
+#define MAX(a,b) (((a)>(b))?(a):(b))
+#endif
 
 /* types ==================================================================== */
 typedef enum {
@@ -80,7 +91,9 @@ typedef enum {
 
 typedef enum {
   eFormatDec,
+  eFormatInt16,
   eFormatHex,
+  eFormatString,
   eFormatInt,
   eFormatFloat,
   eFormatBin,
@@ -135,21 +148,29 @@ static const int iStopbitsList[] = {
 };
 #ifdef MBPOLL_FLOAT_DISABLE
 static const char * sFormatList[] = {
+  "int16",
   "hex",
+  "string",
   "int"
 };
 static const int iFormatList[] = {
+  eFormatInt16,
   eFormatHex,
+  eFormatString,
   eFormatInt
 };
 #else
 static const char * sFormatList[] = {
+  "int16",
   "hex",
+  "string",
   "int",
   "float"
 };
 static const int iFormatList[] = {
+  eFormatInt16,
   eFormatHex,
+  eFormatString,
   eFormatInt,
   eFormatFloat
 };
@@ -204,7 +225,8 @@ typedef struct xMbPollContext {
   eFormats eFormat;
   int * piSlaveAddr;
   int iSlaveCount;
-  int iStartRef;
+  int * piStartRef;
+  int iStartCount;
   int iCount;
   int iPollRate;
   double dTimeout;
@@ -222,6 +244,7 @@ typedef struct xMbPollContext {
   bool bIsReportSlaveID;
   bool bIsDefaultMode;
   int iPduOffset;
+  bool bWriteSingleAsMany;
   bool bIsChipIo;
   bool bIsBigEndian;
   bool bIsQuiet;
@@ -248,7 +271,8 @@ static xMbPollContext ctx = {
   .eFormat = eFormatDec,
   .piSlaveAddr = NULL,
   .iSlaveCount = -1,
-  .iStartRef = DEFAULT_STARTREF,
+  .piStartRef = NULL,
+  .iStartCount = -1,
   .iCount = DEFAULT_NUMOFVALUES,
   .iPollRate = DEFAULT_POLLRATE,
   .dTimeout = DEFAULT_TIMEOUT,
@@ -268,6 +292,7 @@ static xMbPollContext ctx = {
   .bIsWrite = true,
   .bIsDefaultMode = true,
   .iPduOffset = 1,
+  .bWriteSingleAsMany = false,
   .bIsChipIo = false,
   .bIsBigEndian = false,
   .bIsQuiet = false,
@@ -299,14 +324,14 @@ static xChipIoSerial * xChipSerial;
 static const char sChipIoSlaveAddrStr[] = "chipio slave address";
 static const char sChipIoIrqPinStr[] = "chipio irq pin";
 // option -i et -n supplémentaires pour chipio
-static const char * short_options = "m:a:r:c:t:1l:o:p:b:d:s:P:u0RhVvwBqi:n:";
+static const char * short_options = "m:a:r:c:t:1l:o:p:b:d:s:P:u0WRhVvwBqi:n:";
 
 #else /* USE_CHIPIO == 0 */
 /* constants ================================================================ */
 #ifdef MBPOLL_GPIO_RTS
-static const char * short_options = "m:a:r:c:t:1l:o:p:b:d:s:P:u0R::F::hVvwBq";
+static const char * short_options = "m:a:r:c:t:1l:o:p:b:d:s:P:u0WR::F::hVvwBq";
 #else
-static const char * short_options = "m:a:r:c:t:1l:o:p:b:d:s:P:u0RFhVvwBq";
+static const char * short_options = "m:a:r:c:t:1l:o:p:b:d:s:P:u0WRFhVvwBq";
 #endif
 // -----------------------------------------------------------------------------
 #endif /* USE_CHIPIO == 0 */
@@ -342,25 +367,46 @@ float fSwapFloat (float f);
 int32_t lSwapLong (int32_t l);
 void mb_delay (unsigned long d);
 
-#if defined(_MSC_VER)
-// Portage des fonctions POSIX ou GNU
+#if defined (__unix__) || (defined (__APPLE__) && defined (__MACH__))
+// Portage des fonctions Microsoft
 // -----------------------------------------------------------------------------
 static char *
+_strlwr (char * str) {
+  char * p = str;
+
+  while (*p) {
+    *p = tolower (*p);
+    p++;
+  }
+  return str;
+}
+#else
+// Portage des fonctions POSIX ou GNU
+
+// -----------------------------------------------------------------------------
+// posix
+#define strcasecmp _stricmp
+
+// -----------------------------------------------------------------------------
+// posix
+static char *
 basename (char * path) {
   static char fname[_MAX_FNAME];
   _splitpath (path, NULL, NULL, fname, NULL);
 
   return fname;
 }
+
 // -----------------------------------------------------------------------------
+// posix
 static char *
 strcasestr (const char *haystack, const char *needle) {
-  int nlen = strlen (needle);
-  int hlen = strlen (haystack) - nlen + 1;
-  int i;
+  size_t nlen = strlen (needle);
+  size_t hlen = strlen (haystack) - nlen + 1;
+  size_t i;
 
   for (i = 0; i < hlen; i++) {
-    int j;
+    size_t j;
     for (j = 0; j < nlen; j++) {
       unsigned char c1 = haystack[i + j];
       unsigned char c2 = needle[j];
@@ -376,6 +422,7 @@ next:
 }
 
 // -----------------------------------------------------------------------------
+// posix
 static char *
 index (const char *s, int c) {
 
@@ -387,22 +434,8 @@ index (const char *s, int c) {
   }
   return NULL;
 }
-#else /* _MSC_VER not defined */
-// Portage des fonctions Microsoft
 
-// -----------------------------------------------------------------------------
-static char *
-strlwr (char * str) {
-  char * p = str;
-
-  while (*p) {
-    *p = tolower (*p);
-    p++;
-  }
-  return str;
-}
-
-#endif /* _MSC_VER not defined */
+#endif
 
 /* main ===================================================================== */
 
@@ -436,7 +469,7 @@ main (int argc, char **argv) {
         break;
 
       case 'r':
-        ctx.iStartRef = iGetInt (sStartRefStr, optarg, 0);
+        ctx.piStartRef = iGetIntList (sStartRefStr, optarg, &ctx.iStartCount);
         break;
 
       case 'c':
@@ -491,6 +524,10 @@ main (int argc, char **argv) {
         ctx.iPduOffset = 0;
         break;
 
+      case 'W':
+        ctx.bWriteSingleAsMany = true;
+        break;
+
       case 'l':
         ctx.iPollRate = iGetInt (sPollRateStr, optarg, 0);
         if (ctx.iPollRate < POLLRATE_MIN) {
@@ -569,15 +606,30 @@ main (int argc, char **argv) {
   }
   while (iNextOption != -1);
 
-  if (ctx.iPduOffset) {
+  if (ctx.iStartCount == -1) {
+    ctx.piStartRef = malloc (sizeof (int));
+    assert (ctx.piStartRef);
+    ctx.piStartRef[0] = DEFAULT_STARTREF;
+    ctx.iStartCount = 1;
+  }
 
-    vCheckIntRange (sStartRefStr, ctx.iStartRef,
-                    STARTREF_MIN, STARTREF_MAX);
+  int i;
+  if (ctx.iPduOffset) {
+    for (i = 0; i < ctx.iStartCount; i++) {
+      vCheckIntRange (sStartRefStr, ctx.piStartRef[i],
+                      STARTREF_MIN, STARTREF_MAX);
+    }
   }
   else {
+    for (i = 0; i < ctx.iStartCount; i++) {
+      vCheckIntRange (sStartRefStr, ctx.piStartRef[i],
+                      STARTREF_MIN - 1, STARTREF_MAX - 1);
+    }
+  }
 
-    vCheckIntRange (sStartRefStr, ctx.iStartRef,
-                    STARTREF_MIN - 1, STARTREF_MAX - 1);
+  // ignore iCount > 1 if start ref list contains more then one value
+  if ((ctx.iStartCount > 1) && (ctx.iCount > 1)) {
+    ctx.iCount = 1;
   }
 
   // Coils et Discrete inputs toujours en binaire
@@ -701,6 +753,15 @@ main (int argc, char **argv) {
               DFLOAT (ctx.pvData, i) = fSwapFloat ( (float) dValue);
               PDEBUG ("Float[%d]=%g\n", i, fSwapFloat (DFLOAT (ctx.pvData, i)));
             }
+            else if (ctx.eFormat == eFormatString) {
+                vSyntaxErrorExit ("You can use string format only for output");
+            }
+            else if (ctx.eFormat == eFormatInt16) {
+              iValue = iGetInt (sDataStr, argv[arg], 0);
+              vCheckIntRange (sDataStr, iValue, INT16_MIN, INT16_MAX);
+              DUINT16 (ctx.pvData, i) = (uint16_t) iValue;
+              PDEBUG ("Word[%d]=0x%X\n", i, DUINT16 (ctx.pvData, i));
+            }
             else {
               iValue = iGetInt (sDataStr, argv[arg], 0);
               vCheckIntRange (sDataStr, iValue, 0, UINT16_MAX);
@@ -720,6 +781,10 @@ main (int argc, char **argv) {
     vSyntaxErrorExit ("You can give a slave address list only for reading");
   }
 
+  if ( (ctx.iStartCount > 1) && (ctx.bIsWrite) ) {
+    vSyntaxErrorExit ("You can give a start ref list only for reading");
+  }
+
   if (ctx.iSlaveCount == -1) {
 
     ctx.piSlaveAddr = malloc (sizeof (int));
@@ -727,7 +792,6 @@ main (int argc, char **argv) {
     ctx.piSlaveAddr[0] = DEFAULT_SLAVEADDR;
     ctx.iSlaveCount = 1;
   }
-  int i;
 
   // Fin de vérification des valeurs de paramètres et création des contextes
   switch (ctx.eMode) {
@@ -824,14 +888,15 @@ main (int argc, char **argv) {
     // int32 et float utilisent 2 registres 16 bits
     iNbReg = ( (ctx.eFormat == eFormatInt) || (ctx.eFormat == eFormatFloat)) ?
              ctx.iCount * 2 : ctx.iCount;
-    // libmodbus utilise les adresses PDU !
-    iStartReg = ctx.iStartRef - ctx.iPduOffset;
 
     // Début de la boucle de scrutation
     do {
 
       if (ctx.bIsWrite) {
 
+        // libmodbus utilise les adresses PDU !
+        iStartReg = ctx.piStartRef[0] - ctx.iPduOffset;
+
         modbus_set_slave (ctx.xBus, ctx.piSlaveAddr[0]);
         ctx.iTxCount++;
 
@@ -853,7 +918,7 @@ main (int argc, char **argv) {
             break;
 
           case eFuncHoldingReg:
-            if (iNbReg == 1) {
+            if (iNbReg == 1 && (!ctx.bWriteSingleAsMany)) {
 
               // Ecriture d'un seul registre
               iRet = modbus_write_register (ctx.xBus, iStartReg,
@@ -900,41 +965,47 @@ main (int argc, char **argv) {
             putchar ('\n');
           }
 
-          switch (ctx.eFunction) {
-            case eFuncDiscreteInput:
-              iRet = modbus_read_input_bits (ctx.xBus, iStartReg, iNbReg,
-                                             ctx.pvData);
-              break;
+          int j;
+          for (j = 0; j < ctx.iStartCount; j++) {
+            // libmodbus utilise les adresses PDU !
+            iStartReg = ctx.piStartRef[j] - ctx.iPduOffset;
 
-            case eFuncCoil:
-              iRet = modbus_read_bits (ctx.xBus, iStartReg, iNbReg,
-                                       ctx.pvData);
-              break;
+            switch (ctx.eFunction) {
+              case eFuncDiscreteInput:
+                iRet = modbus_read_input_bits (ctx.xBus, iStartReg, iNbReg,
+                                               ctx.pvData);
+                break;
 
-            case eFuncInputReg:
-              iRet = modbus_read_input_registers (ctx.xBus, iStartReg, iNbReg,
-                                                  ctx.pvData);
-              break;
+              case eFuncCoil:
+                iRet = modbus_read_bits (ctx.xBus, iStartReg, iNbReg,
+                                         ctx.pvData);
+                break;
 
-            case eFuncHoldingReg:
-              iRet = modbus_read_registers (ctx.xBus, iStartReg, iNbReg,
-                                            ctx.pvData);
-              break;
+              case eFuncInputReg:
+                iRet = modbus_read_input_registers (ctx.xBus, iStartReg, iNbReg,
+                                                    ctx.pvData);
+                break;
 
-            default: // Impossible, la valeur a été vérifiée, évite un warning de gcc
-              break;
+              case eFuncHoldingReg:
+                iRet = modbus_read_registers (ctx.xBus, iStartReg, iNbReg,
+                                              ctx.pvData);
+                break;
 
-          }
-          if (iRet == iNbReg) {
+              default: // Impossible, la valeur a été vérifiée, évite un warning de gcc
+                break;
 
-            ctx.iRxCount++;
-            vPrintReadValues (ctx.iStartRef, ctx.iCount, &ctx);
-          }
-          else {
-            ctx.iErrorCount++;
-            fprintf (stderr, "Read %s failed: %s\n",
-                     sFunctionToStr (ctx.eFunction),
-                     modbus_strerror (errno));
+            }
+            if (iRet == iNbReg) {
+
+              ctx.iRxCount++;
+              vPrintReadValues (ctx.piStartRef[j], ctx.iCount, &ctx);
+            }
+            else {
+              ctx.iErrorCount++;
+              fprintf (stderr, "Read %s failed: %s\n",
+                       sFunctionToStr (ctx.eFunction),
+                       modbus_strerror (errno));
+            }
           }
           if (ctx.bIsPolling) {
 
@@ -984,11 +1055,21 @@ vPrintReadValues (int iAddr, int iCount, xMbPollContext * ctx) {
       }
       break;
 
+      case eFormatInt16:
+        printf ("%d", (int) (int16_t) (DUINT16 (ctx->pvData, i)));
+        iAddr++;
+        break;
+
       case eFormatHex:
         printf ("0x%04X", DUINT16 (ctx->pvData, i));
         iAddr++;
         break;
 
+      case eFormatString:
+        printf ("%c%c", (char) ((int) (DUINT16 (ctx->pvData, i) / 256)), (char) (DUINT16 (ctx->pvData, i) % 256));
+        iAddr++;
+        break;
+
       case eFormatInt:
         printf ("%d", lSwapLong (DINT32 (ctx->pvData, i)));
         iAddr += 2;
@@ -1013,7 +1094,7 @@ vReportSlaveID (const xMbPollContext * ctx) {
 
   modbus_set_slave (ctx->xBus, ctx->piSlaveAddr[0]);
   // Affichage de la configuration
-  printf ("Protocol configuration: Modbus %s\n", sModeList[ctx->eMode]);
+  printf ("Protocol configuration: ModBus %s\n", sModeList[ctx->eMode]);
   printf ("Slave configuration...: address = %d, report slave id\n",
           ctx->piSlaveAddr[0]);
 
@@ -1108,11 +1189,18 @@ void
 vPrintConfig (const xMbPollContext * ctx) {
 
   // Affichage de la configuration
-  printf ("Protocol configuration: Modbus %s\n", sModeList[ctx->eMode]);
+  printf ("Protocol configuration: ModBus %s\n", sModeList[ctx->eMode]);
   printf ("Slave configuration...: address = ");
   vPrintIntList (ctx->piSlaveAddr, ctx->iSlaveCount);
-  printf ("\n                        start reference = %d, count = %d\n",
-          ctx->iStartRef, ctx->iCount);
+  if (ctx->iStartCount > 1) {
+    printf ("\n                        start reference = ");
+    vPrintIntList (ctx->piStartRef, ctx->iStartCount);
+    printf ("\n");
+  }
+  else {
+    printf ("\n                        start reference = %d, count = %d\n",
+            ctx->piStartRef[0], ctx->iCount);
+  }
   vPrintCommunicationSetup (ctx);
   printf ("Data type.............: ");
   switch (ctx->eFunction) {
@@ -1257,7 +1345,7 @@ vVersion (void)  {
 void
 vWarranty (void) {
   printf (
-    "Copyright © 2015-2019 %s, All rights reserved.\n\n"
+    "Copyright (c) 2015-2023 %s, All rights reserved.\n\n"
 
     " mbpoll is free software: you can redistribute it and/or modify\n"
     " it under the terms of the GNU General Public License as published by\n"
@@ -1278,9 +1366,9 @@ vWarranty (void) {
 // -----------------------------------------------------------------------------
 void
 vHello (void) {
-  printf ("mbpoll %s - FieldTalk(tm) Modbus(R) Master Simulator\n",
+  printf ("mbpoll %s - ModBus(R) Master Simulator\n",
           VERSION_SHORT);
-  printf ("Copyright © 2015-2019 %s, %s\n", AUTHORS, WEBSITE);
+  printf ("Copyright (c) 2015-2023 %s, %s\n", AUTHORS, WEBSITE);
   printf ("This program comes with ABSOLUTELY NO WARRANTY.\n");
   printf ("This is free software, and you are welcome to redistribute it\n");
   printf ("under certain conditions; type 'mbpoll -w' for details.\n\n");
@@ -1295,8 +1383,10 @@ vUsage (FILE * stream, int exit_msg) {
            , sMyName);
 
   fprintf (stream,
-           "ModBus Master Simulator. It allows to read and write in ModBus slave registers\n"
-           "                         connected by serial (RTU only) or TCP.\n\n"
+          //-----------------------------------------------------------------------------
+           "ModBus(R) Master Simulator Master Simulator.\n"
+           " Allows to read and write in ModBus slave registers connected by\n"
+           " serial (RTU only) or TCP.\n\n"
 
            "Arguments :\n"
            "  device        Serial port when using ModBus RTU protocol\n"
@@ -1328,28 +1418,36 @@ vUsage (FILE * stream, int exit_msg) {
            "                separated by commas or colons, for example :\n"
            "                -a 32,33,34,36:40 read [32,33,34,36,37,38,39,40]\n"
            "  -r #          Start reference (%d is default)\n"
+           "                for reading, it is possible to give a reference list\n"
+           "                separated by commas or colons\n"
            "  -c #          Number of values to read (%d-%d, %d is default)\n"
            "  -u            Read the description of the type, the current status, and other\n"
            "                information specific to a remote device (RTU only)\n"
            "  -t 0          Discrete output (coil) data type (binary 0 or 1)\n"
            "  -t 1          Discrete input data type (binary 0 or 1)\n"
            "  -t 3          16-bit input register data type\n"
+           "  -t 3:int16    16-bit input register data type with signed int display\n"
            "  -t 3:hex      16-bit input register data type with hex display\n"
+           "  -t 3:string   16-bit input register data type with string (char) display\n"
            "  -t 3:int      32-bit integer data type in input register table\n"
 #ifndef MBPOLL_FLOAT_DISABLE
            "  -t 3:float    32-bit float data type in input register table\n"
 #endif
            "  -t 4          16-bit output (holding) register data type (default)\n"
+           "  -t 4:int16    16-bit output (holding) register data type with signed int display\n"
            "  -t 4:hex      16-bit output (holding) register data type with hex display\n"
+           "  -t 4:string   16-bit output (holding) register data type with string (char) display\n"
            "  -t 4:int      32-bit integer data type in output (holding) register table\n"
 #ifndef MBPOLL_FLOAT_DISABLE
            "  -t 4:float    32-bit float data type in output (holding) register table\n"
 #endif
            "  -0            First reference is 0 (PDU addressing) instead 1\n"
+           "  -W            Using function 10 for write a single register\n"
            "  -B            Big endian word order for 32-bit integer and float\n"
            "  -1            Poll only once only, otherwise every poll rate interval\n"
            "  -l #          Poll rate in ms, ( > %d, %d is default)\n"
            "  -o #          Time-out in seconds (%.2f - %.2f, %.2f s is default)\n"
+           "  -q            Quiet mode.  Minimum output only\n"
            "Options for ModBus / TCP : \n"
            "  -p #          TCP port number (%s is default)\n"
            "Options for ModBus RTU : \n"
@@ -1455,7 +1553,7 @@ iGetEnum (const char * sName, char * sElmt, const char ** psStrList,
 
     if (strcasecmp (sElmt, psStrList[i]) == 0) {
 
-      PDEBUG ("Set %s=%s\n", sName, strlwr (sElmt));
+      PDEBUG ("Set %s=%s\n", sName, _strlwr (sElmt));
       return iList[i];
     }
   }
@@ -1693,7 +1791,7 @@ void
 mb_delay (unsigned long d) {
 
   if (d) {
-
+#if defined (__unix__) || (defined (__APPLE__) && defined (__MACH__))
     if (d == -1) {
 
       sleep (-1);
@@ -1705,6 +1803,9 @@ mb_delay (unsigned long d) {
       dt.tv_sec  = d / 1000UL;
       nanosleep (&dt, NULL);
     }
+#else
+    Sleep (d);
+#endif
   }
 }
 /* ========================================================================== */
